@@ -4,6 +4,7 @@ package capi
 
 import (
 	"bytes"
+	"cmp"
 	"context"
 	"crypto"
 	"crypto/ecdsa"
@@ -15,41 +16,59 @@ import (
 	"fmt"
 	"io"
 	"math/big"
+	"net/url"
 	"reflect"
+	"strconv"
 	"strings"
 	"unsafe"
 
 	"github.com/pkg/errors"
-	"go.step.sm/crypto/kms/apiv1"
-	"go.step.sm/crypto/kms/uri"
-	"go.step.sm/crypto/randutil"
+
 	"golang.org/x/crypto/cryptobyte"
 	"golang.org/x/crypto/cryptobyte/asn1"
 	"golang.org/x/sys/windows"
+
+	"go.step.sm/crypto/fingerprint"
+	"go.step.sm/crypto/kms/apiv1"
+	"go.step.sm/crypto/kms/uri"
+	"go.step.sm/crypto/randutil"
 )
 
 // Scheme is the scheme used in uris, the string "capi".
 const Scheme = string(apiv1.CAPIKMS)
 
 const (
-	ProviderNameArg        = "provider"
-	ContainerNameArg       = "key"
-	HashArg                = "sha1"
-	StoreLocationArg       = "store-location" // 'machine', 'user', etc
-	StoreNameArg           = "store"          // 'MY', 'CA', 'ROOT', etc
-	KeyIDArg               = "key-id"
-	SubjectCNArg           = "cn"
-	SerialNumberArg        = "serial"
-	IssuerNameArg          = "issuer"
-	KeySpec                = "key-spec"                  // 0, 1, 2; none/NONE, at_keyexchange/AT_KEYEXCHANGE, at_signature/AT_SIGNATURE
-	SkipFindCertificateKey = "skip-find-certificate-key" // skips looking up certificate private key when storing a certificate
+	ProviderNameArg              = "provider"
+	ContainerNameArg             = "key"
+	HashArg                      = "sha1"
+	StoreLocationArg             = "store-location" // 'machine', 'user', etc
+	StoreNameArg                 = "store"          // 'MY', 'CA', 'ROOT', etc
+	IntermediateStoreLocationArg = "intermediate-store-location"
+	IntermediateStoreNameArg     = "intermediate-store"
+	KeyIDArg                     = "key-id"
+	SubjectCNArg                 = "cn"
+	SerialNumberArg              = "serial"
+	IssuerNameArg                = "issuer"
+	KeySpec                      = "key-spec"                  // 0, 1, 2; none/NONE, at_keyexchange/AT_KEYEXCHANGE, at_signature/AT_SIGNATURE
+	SkipFindCertificateKey       = "skip-find-certificate-key" // skips looking up certificate private key when storing a certificate
 )
 
 const (
-	MachineStore = "machine"
-	UserStore    = "user"
+	MachineStoreLocation = "machine"
+	UserStoreLocation    = "user"
+	MyStore              = "My"
+	CAStore              = "CA" // TODO(hs): verify "CA" works for "machine" certs too
+
+	// Deprecated: use MachineStoreLocation
+	MachineStore = MachineStoreLocation
+	// Deprecated: use UserStoreLocation
+	UserStore = UserStoreLocation
 )
 
+// maximumIterations is the maximum number of times for the recursive
+// intermediate CA lookup loop.
+const maximumIterations = 10
+
 var signatureAlgorithmMapping = map[apiv1.SignatureAlgorithm]string{
 	apiv1.UnspecifiedSignAlgorithm: ALG_ECDSA_P256,
 	apiv1.SHA256WithRSA:            ALG_RSA,
@@ -60,6 +79,59 @@ var signatureAlgorithmMapping = map[apiv1.SignatureAlgorithm]string{
 	apiv1.ECDSAWithSHA512:          ALG_ECDSA_P521,
 }
 
+type uriAttributes struct {
+	containerName             string
+	hash                      []byte
+	storeLocation             string
+	storeName                 string
+	intermediateStoreLocation string
+	intermediateStoreName     string
+	keyID                     []byte
+	subjectCN                 string
+	serialNumber              string
+	issuerName                string
+	keySpec                   string
+	skipFindCertificateKey    bool
+	pin                       string
+}
+
+func parseURI(rawuri string) (*uriAttributes, error) {
+	u, err := uri.ParseWithScheme(Scheme, rawuri)
+	if err != nil {
+		return nil, fmt.Errorf("failed to parse URI: %w", err)
+	}
+
+	var hashValue []byte
+	if u.Has(HashArg) {
+		if hashValue, err = u.GetHexEncoded(HashArg); err != nil {
+			return nil, fmt.Errorf("failed getting %s from URI %q: %w", HashArg, rawuri, err)
+		}
+	}
+
+	var keyIDValue []byte
+	if u.Has(KeyIDArg) {
+		if keyIDValue, err = u.GetHexEncoded(KeyIDArg); err != nil {
+			return nil, fmt.Errorf("failed getting %s from URI %q: %w", KeyIDArg, rawuri, err)
+		}
+	}
+
+	return &uriAttributes{
+		containerName:             u.Get(ContainerNameArg),
+		hash:                      hashValue,
+		storeLocation:             cmp.Or(u.Get(StoreLocationArg), UserStoreLocation),
+		storeName:                 cmp.Or(u.Get(StoreNameArg), MyStore),
+		intermediateStoreLocation: cmp.Or(u.Get(IntermediateStoreLocationArg), UserStoreLocation),
+		intermediateStoreName:     cmp.Or(u.Get(IntermediateStoreNameArg), CAStore),
+		keyID:                     keyIDValue,
+		subjectCN:                 u.Get(SubjectCNArg),
+		serialNumber:              u.Get(SerialNumberArg),
+		issuerName:                u.Get(IssuerNameArg),
+		keySpec:                   u.Get(KeySpec),
+		skipFindCertificateKey:    u.GetBool(SkipFindCertificateKey),
+		pin:                       u.Pin(),
+	}, nil
+}
+
 // CAPIKMS implements a KMS using Windows CryptoAPI (CAPI) and Next-Gen CryptoAPI (CNG).
 //
 // The URI format used in CAPIKMS is the following:
@@ -293,49 +365,20 @@ func (k *CAPIKMS) Close() error {
 
 // getCertContext returns a pointer to a X.509 certificate context based on the provided URI
 // callers are responsible for freeing the context
-func (k *CAPIKMS) getCertContext(req *apiv1.LoadCertificateRequest) (*windows.CertContext, error) {
-	u, err := uri.ParseWithScheme(Scheme, req.Name)
-	if err != nil {
-		return nil, fmt.Errorf("failed to parse URI: %w", err)
-	}
-
-	sha1Hash, err := u.GetHexEncoded(HashArg)
-	switch {
-	case err != nil:
-		return nil, fmt.Errorf("failed getting %s from URI %q: %w", HashArg, req.Name, err)
-	case len(sha1Hash) > 0 && len(sha1Hash) != 20:
-		return nil, fmt.Errorf("decoded %s has length %d; expected 20 bytes for SHA-1", HashArg, len(sha1Hash))
-	}
-
-	keyID, err := u.GetHexEncoded(KeyIDArg)
-	if err != nil {
-		return nil, fmt.Errorf("failed getting %s from URI %q: %w", KeyIDArg, req.Name, err)
-	}
-	issuerName := u.Get(IssuerNameArg)
-	subjectCN := u.Get(SubjectCNArg)
-	serialNumber := u.Get(SerialNumberArg)
-
-	// default to the user store
-	var storeLocation string
-	if storeLocation = u.Get(StoreLocationArg); storeLocation == "" {
-		storeLocation = UserStore
+func (k *CAPIKMS) getCertContext(u *uriAttributes) (*windows.CertContext, error) {
+	// The hash argument is a SHA-1
+	if len(u.hash) > 0 && len(u.hash) != 20 {
+		return nil, fmt.Errorf("decoded %s has length %d; expected 20 bytes for SHA-1", HashArg, len(u.hash))
 	}
 
 	var certStoreLocation uint32
-	switch storeLocation {
-	case UserStore:
+	switch u.storeLocation {
+	case UserStoreLocation:
 		certStoreLocation = certStoreCurrentUser
-	case MachineStore:
+	case MachineStoreLocation:
 		certStoreLocation = certStoreLocalMachine
 	default:
-		return nil, fmt.Errorf("invalid cert store location %q", storeLocation)
-	}
-
-	var storeName string
-
-	// default to the 'My' store
-	if storeName = u.Get(StoreNameArg); storeName == "" {
-		storeName = "My"
+		return nil, fmt.Errorf("invalid cert store location %q", u.storeLocation)
 	}
 
 	st, err := windows.CertOpenStore(
@@ -343,20 +386,20 @@ func (k *CAPIKMS) getCertContext(req *apiv1.LoadCertificateRequest) (*windows.Ce
 		0,
 		0,
 		certStoreLocation,
-		uintptr(unsafe.Pointer(wide(storeName))))
+		uintptr(unsafe.Pointer(wide(u.storeName))))
 	if err != nil {
-		return nil, fmt.Errorf("CertOpenStore for the %q store %q returned: %w", storeLocation, storeName, err)
+		return nil, fmt.Errorf("CertOpenStore for the %q store %q returned: %w", u.storeLocation, u.storeName, err)
 	}
 
 	var handle *windows.CertContext
 
 	switch {
-	case len(sha1Hash) > 0:
+	case len(u.hash) > 0:
 		searchData := CERT_ID_KEYIDORHASH{
 			idChoice: CERT_ID_SHA1_HASH,
 			KeyIDOrHash: CRYPTOAPI_BLOB{
-				len:  uint32(len(sha1Hash)),
-				data: uintptr(unsafe.Pointer(&sha1Hash[0])),
+				len:  uint32(len(u.hash)),
+				data: uintptr(unsafe.Pointer(&u.hash[0])),
 			},
 		}
 		handle, err = findCertificateInStore(st,
@@ -368,14 +411,14 @@ func (k *CAPIKMS) getCertContext(req *apiv1.LoadCertificateRequest) (*windows.Ce
 			return nil, fmt.Errorf("findCertificateInStore failed: %w", err)
 		}
 		if handle == nil {
-			return nil, apiv1.NotFoundError{Message: fmt.Sprintf("certificate with %s=%s not found", HashArg, u.Get(HashArg))}
+			return nil, apiv1.NotFoundError{Message: fmt.Sprintf("certificate with %s=%x not found", HashArg, u.hash)}
 		}
-	case len(keyID) > 0:
+	case len(u.keyID) > 0:
 		searchData := CERT_ID_KEYIDORHASH{
 			idChoice: CERT_ID_KEY_IDENTIFIER,
 			KeyIDOrHash: CRYPTOAPI_BLOB{
-				len:  uint32(len(keyID)),
-				data: uintptr(unsafe.Pointer(&keyID[0])),
+				len:  uint32(len(u.keyID)),
+				data: uintptr(unsafe.Pointer(&u.keyID[0])),
 			},
 		}
 		handle, err = findCertificateInStore(st,
@@ -387,22 +430,22 @@ func (k *CAPIKMS) getCertContext(req *apiv1.LoadCertificateRequest) (*windows.Ce
 			return nil, fmt.Errorf("findCertificateInStore failed: %w", err)
 		}
 		if handle == nil {
-			return nil, apiv1.NotFoundError{Message: fmt.Sprintf("certificate with %s=%s not found", KeyIDArg, keyID)}
+			return nil, apiv1.NotFoundError{Message: fmt.Sprintf("certificate with %s=%x not found", KeyIDArg, u.keyID)}
 		}
-	case issuerName != "" && (serialNumber != "" || subjectCN != ""):
+	case u.issuerName != "" && (u.serialNumber != "" || u.subjectCN != ""):
 		var prevCert *windows.CertContext
 		for {
 			handle, err = findCertificateInStore(st,
 				encodingX509ASN|encodingPKCS7,
 				0,
 				findIssuerStr,
-				uintptr(unsafe.Pointer(wide(issuerName))), prevCert)
+				uintptr(unsafe.Pointer(wide(u.issuerName))), prevCert)
 			if err != nil {
 				return nil, fmt.Errorf("findCertificateInStore failed: %w", err)
 			}
 
 			if handle == nil {
-				return nil, apiv1.NotFoundError{Message: fmt.Sprintf("certificate with %s=%q not found", IssuerNameArg, issuerName)}
+				return nil, apiv1.NotFoundError{Message: fmt.Sprintf("certificate with %s=%q not found", IssuerNameArg, u.issuerName)}
 			}
 
 			x509Cert, err := certContextToX509(handle)
@@ -411,13 +454,13 @@ func (k *CAPIKMS) getCertContext(req *apiv1.LoadCertificateRequest) (*windows.Ce
 			}
 
 			switch {
-			case len(serialNumber) > 0:
+			case len(u.serialNumber) > 0:
 				// TODO: Replace this search with a CERT_ID + CERT_ISSUER_SERIAL_NUMBER search instead
 				// https://learn.microsoft.com/en-us/windows/win32/api/wincrypt/ns-wincrypt-cert_id
 				// https://learn.microsoft.com/en-us/windows/win32/api/wincrypt/ns-wincrypt-cert_issuer_serial_number
 				var bi *big.Int
-				if strings.HasPrefix(serialNumber, "0x") {
-					serialBytes, err := hex.DecodeString(strings.TrimPrefix(serialNumber, "0x"))
+				if strings.HasPrefix(u.serialNumber, "0x") {
+					serialBytes, err := hex.DecodeString(strings.TrimPrefix(u.serialNumber, "0x"))
 					if err != nil {
 						return nil, fmt.Errorf("invalid hex format for %s: %w", SerialNumberArg, err)
 					}
@@ -425,7 +468,7 @@ func (k *CAPIKMS) getCertContext(req *apiv1.LoadCertificateRequest) (*windows.Ce
 					bi = new(big.Int).SetBytes(serialBytes)
 				} else {
 					bi := new(big.Int)
-					bi, ok := bi.SetString(serialNumber, 10)
+					bi, ok := bi.SetString(u.serialNumber, 10)
 					if !ok {
 						return nil, fmt.Errorf("invalid %s - must be in hex or integer format", SerialNumberArg)
 					}
@@ -434,8 +477,8 @@ func (k *CAPIKMS) getCertContext(req *apiv1.LoadCertificateRequest) (*windows.Ce
 				if x509Cert.SerialNumber.Cmp(bi) == 0 {
 					return handle, nil
 				}
-			case len(subjectCN) > 0:
-				if x509Cert.Subject.CommonName == subjectCN {
+			case len(u.subjectCN) > 0:
+				if x509Cert.Subject.CommonName == u.subjectCN {
 					return handle, nil
 				}
 			}
@@ -451,31 +494,29 @@ func (k *CAPIKMS) getCertContext(req *apiv1.LoadCertificateRequest) (*windows.Ce
 
 // CreateSigner returns a crypto.Signer that will sign using the key passed in via the URI.
 func (k *CAPIKMS) CreateSigner(req *apiv1.CreateSignerRequest) (crypto.Signer, error) {
-	u, err := uri.ParseWithScheme(Scheme, req.SigningKey)
+	u, err := parseURI(req.SigningKey)
 	if err != nil {
-		return nil, fmt.Errorf("failed to parse URI: %w", err)
+		return nil, err
 	}
 
 	var (
-		kh            uintptr
-		certHandle    *windows.CertContext
-		containerName string
+		kh         uintptr
+		certHandle *windows.CertContext
 	)
-	if containerName = u.Get(ContainerNameArg); containerName != "" {
+
+	if u.containerName != "" {
 		keyFlags, err := k.getKeyFlags(u)
 		if err != nil {
 			return nil, err
 		}
 
-		kh, err = nCryptOpenKey(k.providerHandle, containerName, 0, keyFlags)
+		kh, err = nCryptOpenKey(k.providerHandle, u.containerName, 0, keyFlags)
 		if err != nil {
-			return nil, fmt.Errorf("unable to open key using %q=%q: %w", ContainerNameArg, containerName, err)
+			return nil, fmt.Errorf("unable to open key using %q=%q: %w", ContainerNameArg, u.containerName, err)
 		}
 	} else {
 		// check if a certificate can be located using the URI
-		certHandle, err = k.getCertContext(&apiv1.LoadCertificateRequest{
-			Name: req.SigningKey,
-		})
+		certHandle, err = k.getCertContext(u)
 		if err != nil {
 			return nil, fmt.Errorf("%v not specified", ContainerNameArg)
 		}
@@ -486,18 +527,17 @@ func (k *CAPIKMS) CreateSigner(req *apiv1.CreateSignerRequest) (crypto.Signer, e
 		}
 	}
 
-	pinOrPass := u.Pin()
-	if pinOrPass == "" {
-		pinOrPass = k.pin
+	if u.pin == "" {
+		u.pin = k.pin
 	}
 
-	if pinOrPass != "" && k.providerName == ProviderMSSC {
-		err = nCryptSetProperty(kh, NCRYPT_PIN_PROPERTY, pinOrPass, 0)
+	if u.pin != "" && k.providerName == ProviderMSSC {
+		err = nCryptSetProperty(kh, NCRYPT_PIN_PROPERTY, u.pin, 0)
 		if err != nil {
 			return nil, fmt.Errorf("unable to set key NCRYPT_PIN_PROPERTY: %w", err)
 		}
-	} else if pinOrPass != "" && k.providerName == ProviderMSPCP {
-		passHash, err := hashPasswordUTF16(pinOrPass)
+	} else if u.pin != "" && k.providerName == ProviderMSPCP {
+		passHash, err := hashPasswordUTF16(u.pin)
 		if err != nil {
 			return nil, fmt.Errorf("unable to hash password: %w", err)
 		}
@@ -508,13 +548,12 @@ func (k *CAPIKMS) CreateSigner(req *apiv1.CreateSignerRequest) (crypto.Signer, e
 		}
 	}
 
-	return newCAPISigner(kh, containerName, pinOrPass)
+	return newCAPISigner(kh, u.containerName, u.pin)
 }
 
-func setKeySpec(u *uri.URI) (uint32, error) {
+func setKeySpec(u *uriAttributes) (uint32, error) {
 	keySpec := uint32(0) // default KeySpec value is NONE
-	value := u.Get(KeySpec)
-	if v := strings.ReplaceAll(strings.ToLower(value), "_", ""); v != "" {
+	if v := strings.ReplaceAll(strings.ToLower(u.keySpec), "_", ""); v != "" {
 		switch v {
 		case "0", "none", "null":
 			break // already set as the default
@@ -523,7 +562,7 @@ func setKeySpec(u *uri.URI) (uint32, error) {
 		case "2", "atsignature":
 			keySpec = uint32(2) // AT_SIGNATURE
 		default:
-			return 0, fmt.Errorf("invalid value set for key-spec: %q", value)
+			return 0, fmt.Errorf("invalid value set for key-spec: %q", u.keySpec)
 		}
 	}
 
@@ -542,15 +581,14 @@ func (k *CAPIKMS) CreateKey(req *apiv1.CreateKeyRequest) (*apiv1.CreateKeyRespon
 		return nil, fmt.Errorf("cannot create keys on %s", ProviderMSSC)
 	}
 
-	u, err := uri.ParseWithScheme(Scheme, req.Name)
+	u, err := parseURI(req.Name)
 	if err != nil {
-		return nil, fmt.Errorf("failed to parse URI: %w", err)
+		return nil, err
 	}
 
-	var containerName string
-	if containerName = u.Get(ContainerNameArg); containerName == "" {
-		// generate a uuid for the container name
-		containerName, err = randutil.UUIDv4()
+	// generate a random uuid for the container name if it is not present
+	if u.containerName == "" {
+		u.containerName, err = randutil.UUIDv4()
 		if err != nil {
 			return nil, fmt.Errorf("failed to generate uuid: %w", err)
 		}
@@ -572,7 +610,7 @@ func (k *CAPIKMS) CreateKey(req *apiv1.CreateKeyRequest) (*apiv1.CreateKeyRespon
 	}
 
 	// TODO: check whether RSA keys require legacyKeySpec set to AT_KEYEXCHANGE
-	kh, err := nCryptCreatePersistedKey(k.providerHandle, containerName, alg, keySpec, keyFlags)
+	kh, err := nCryptCreatePersistedKey(k.providerHandle, u.containerName, alg, keySpec, keyFlags)
 	if err != nil {
 		return nil, fmt.Errorf("unable to create persisted key: %w", err)
 	}
@@ -586,22 +624,20 @@ func (k *CAPIKMS) CreateKey(req *apiv1.CreateKeyRequest) (*apiv1.CreateKeyRespon
 		}
 	}
 
-	// if supplied, set the smart card pin/or PCP pass
-	pinOrPass := u.Pin()
-
-	// failover to pin set in kms instantiation
-	if pinOrPass == "" {
-		pinOrPass = k.pin
+	// if supplied, set the smart card pin/or PCP pass, failover to pin set in
+	// kms instantiation
+	if u.pin == "" {
+		u.pin = k.pin
 	}
 
 	// TODO: investigate if there is a similar property for software backed keys
-	if pinOrPass != "" && k.providerName == ProviderMSSC {
-		err = nCryptSetProperty(kh, NCRYPT_PIN_PROPERTY, pinOrPass, 0)
+	if u.pin != "" && k.providerName == ProviderMSSC {
+		err = nCryptSetProperty(kh, NCRYPT_PIN_PROPERTY, u.pin, 0)
 		if err != nil {
 			return nil, fmt.Errorf("unable to set key NCRYPT_PIN_PROPERTY: %w", err)
 		}
-	} else if pinOrPass != "" && k.providerName == ProviderMSPCP {
-		pwHash, err := hashPasswordUTF16(pinOrPass) // we have to SHA1 hash over the utf16 string
+	} else if u.pin != "" && k.providerName == ProviderMSPCP {
+		pwHash, err := hashPasswordUTF16(u.pin) // we have to SHA1 hash over the utf16 string
 		if err != nil {
 			return nil, fmt.Errorf("unable to hash pin: %w", err)
 		}
@@ -639,13 +675,12 @@ func (k *CAPIKMS) CreateKey(req *apiv1.CreateKeyRequest) (*apiv1.CreateKeyRespon
 
 // DeleteKey deletes the key from the key id (Microsoft calls it 'Key Container Name') passed in via the URI
 func (k *CAPIKMS) DeleteKey(req *apiv1.DeleteKeyRequest) error {
-	u, err := uri.ParseWithScheme(Scheme, req.Name)
+	u, err := parseURI(req.Name)
 	if err != nil {
-		return fmt.Errorf("failed to parse URI: %w", err)
+		return err
 	}
 
-	var containerName string
-	if containerName = u.Get(ContainerNameArg); containerName == "" {
+	if u.containerName == "" {
 		return fmt.Errorf("%v not specified", ContainerNameArg)
 	}
 
@@ -654,7 +689,7 @@ func (k *CAPIKMS) DeleteKey(req *apiv1.DeleteKeyRequest) error {
 		return err
 	}
 
-	kh, err := nCryptOpenKey(k.providerHandle, containerName, 0, keyFlags)
+	kh, err := nCryptOpenKey(k.providerHandle, u.containerName, 0, keyFlags)
 	if err != nil {
 		return fmt.Errorf("unable to open key: %w", err)
 	}
@@ -666,13 +701,12 @@ func (k *CAPIKMS) DeleteKey(req *apiv1.DeleteKeyRequest) error {
 
 // GetPublicKey returns the public key from the key id (Microsoft calls it 'Key Container Name') passed in via the URI
 func (k *CAPIKMS) GetPublicKey(req *apiv1.GetPublicKeyRequest) (crypto.PublicKey, error) {
-	u, err := uri.ParseWithScheme(Scheme, req.Name)
+	u, err := parseURI(req.Name)
 	if err != nil {
-		return nil, fmt.Errorf("failed to parse URI: %w", err)
+		return nil, err
 	}
 
-	var containerName string
-	if containerName = u.Get(ContainerNameArg); containerName == "" {
+	if u.containerName == "" {
 		return nil, fmt.Errorf("%v not specified", ContainerNameArg)
 	}
 
@@ -681,7 +715,7 @@ func (k *CAPIKMS) GetPublicKey(req *apiv1.GetPublicKeyRequest) (crypto.PublicKey
 		return nil, err
 	}
 
-	kh, err := nCryptOpenKey(k.providerHandle, containerName, 0, keyFlags)
+	kh, err := nCryptOpenKey(k.providerHandle, u.containerName, 0, keyFlags)
 	if err != nil {
 		return nil, fmt.Errorf("unable to open key: %w", err)
 	}
@@ -694,7 +728,12 @@ func (k *CAPIKMS) GetPublicKey(req *apiv1.GetPublicKeyRequest) (crypto.PublicKey
 // LoadCertificate will return an x509.Certificate if passed a URI containing a subject key
 // identifier (key-id) or sha1 hash
 func (k *CAPIKMS) LoadCertificate(req *apiv1.LoadCertificateRequest) (*x509.Certificate, error) {
-	certHandle, err := k.getCertContext(req)
+	u, err := parseURI(req.Name)
+	if err != nil {
+		return nil, err
+	}
+
+	certHandle, err := k.getCertContext(u)
 	if err != nil {
 		return nil, err
 	}
@@ -703,30 +742,71 @@ func (k *CAPIKMS) LoadCertificate(req *apiv1.LoadCertificateRequest) (*x509.Cert
 	return certContextToX509(certHandle)
 }
 
-func (k *CAPIKMS) StoreCertificate(req *apiv1.StoreCertificateRequest) error {
-	u, err := uri.ParseWithScheme(Scheme, req.Name)
+func (k *CAPIKMS) LoadCertificateChain(req *apiv1.LoadCertificateChainRequest) ([]*x509.Certificate, error) {
+	u, err := parseURI(req.Name)
 	if err != nil {
-		return fmt.Errorf("failed to parse URI: %w", err)
+		return nil, err
 	}
 
-	var storeLocation string
-	if storeLocation = u.Get(StoreLocationArg); storeLocation == "" {
-		storeLocation = UserStore
+	cert, err := k.LoadCertificate(&apiv1.LoadCertificateRequest{
+		Name: req.Name,
+	})
+	if err != nil {
+		return nil, err
+	}
+
+	chain := []*x509.Certificate{cert}
+	child := cert
+	for i := 0; i < maximumIterations; i++ { // loop a maximum number of times
+		authorityKeyID := hex.EncodeToString(child.AuthorityKeyId)
+		parent, err := k.LoadCertificate(&apiv1.LoadCertificateRequest{
+			Name: uri.New(Scheme, url.Values{
+				KeyIDArg:         []string{authorityKeyID},
+				StoreLocationArg: []string{u.intermediateStoreLocation},
+				StoreNameArg:     []string{u.intermediateStoreName},
+			}).String(),
+		})
+		if err != nil {
+			if errors.Is(err, apiv1.NotFoundError{}) {
+				// if error indicates the parent wasn't found, assume end of chain for a specific
+				// combination of store location and store is reached, and break from the loop
+				break
+			}
+			return nil, fmt.Errorf("failed loading intermediate CA certificate using Windows platform cryptography provider: %w", err)
+		}
+
+		// if the discovered parent has a signature from itself, assume it's a root CA,
+		// and break from the loop
+		if parent.CheckSignatureFrom(parent) == nil {
+			break
+		}
+
+		// ensure child has a valid signature from the parent
+		if err := child.CheckSignatureFrom(parent); err != nil {
+			return nil, fmt.Errorf("failed loading intermediate CA certificate using Windows platform cryptography provider: %w", err)
+		}
+
+		chain = append(chain, parent)
+		child = parent
+	}
+
+	return chain, nil
+}
+
+func (k *CAPIKMS) StoreCertificate(req *apiv1.StoreCertificateRequest) error {
+	u, err := parseURI(req.Name)
+	if err != nil {
+		return err
 	}
 
 	var certStoreLocation uint32
-	switch storeLocation {
-	case UserStore:
+	switch u.storeLocation {
+	case UserStoreLocation:
 		certStoreLocation = certStoreCurrentUser
-	case MachineStore:
+	case MachineStoreLocation:
 		certStoreLocation = certStoreLocalMachine
 	default:
-		return fmt.Errorf("invalid cert store location %q", storeLocation)
-	}
-
-	var storeName string
-	if storeName = u.Get(StoreNameArg); storeName == "" {
-		storeName = "My"
+		return fmt.Errorf("invalid cert store location %q", u.storeLocation)
 	}
 
 	certContext, err := windows.CertCreateCertificateContext(
@@ -742,7 +822,7 @@ func (k *CAPIKMS) StoreCertificate(req *apiv1.StoreCertificateRequest) error {
 	// so that looking up the private key for e.g. intermediate certificates can be skipped.
 	// If not skipped, looking up a private key can prompt the user to insert/select a smart
 	// card, which is usually not what we want to happen.
-	if !u.GetBool(SkipFindCertificateKey) {
+	if !u.skipFindCertificateKey {
 		// TODO: not finding the associated private key is not a dealbreaker, but maybe a warning should be issued
 		cryptFindCertificateKeyProvInfo(certContext)
 	}
@@ -752,9 +832,9 @@ func (k *CAPIKMS) StoreCertificate(req *apiv1.StoreCertificateRequest) error {
 		0,
 		0,
 		certStoreLocation,
-		uintptr(unsafe.Pointer(wide(storeName))))
+		uintptr(unsafe.Pointer(wide(u.storeName))))
 	if err != nil {
-		return fmt.Errorf("CertOpenStore for the %q store %q returned: %w", storeLocation, storeName, err)
+		return fmt.Errorf("CertOpenStore for the %q store %q returned: %w", u.storeLocation, u.storeName, err)
 	}
 
 	// Add the cert context to the system certificate store
@@ -765,6 +845,60 @@ func (k *CAPIKMS) StoreCertificate(req *apiv1.StoreCertificateRequest) error {
 	return nil
 }
 
+func (k *CAPIKMS) StoreCertificateChain(req *apiv1.StoreCertificateChainRequest) error {
+	u, err := parseURI(req.Name)
+	if err != nil {
+		return err
+	}
+
+	leaf := req.CertificateChain[0]
+	fp, err := fingerprint.New(leaf.Raw, crypto.SHA1, fingerprint.HexFingerprint)
+	if err != nil {
+		return fmt.Errorf("failed calculating certificate SHA1 fingerprint: %w", err)
+	}
+
+	if err := k.StoreCertificate(&apiv1.StoreCertificateRequest{
+		Name: uri.New("capi", url.Values{
+			HashArg:                []string{fp},
+			StoreLocationArg:       []string{u.storeLocation},
+			StoreNameArg:           []string{u.storeName},
+			SkipFindCertificateKey: []string{strconv.FormatBool(u.skipFindCertificateKey)},
+		}).String(),
+		Certificate: leaf,
+	}); err != nil {
+		return fmt.Errorf("failed storing certificate using Windows platform cryptography provider: %w", err)
+	}
+
+	if len(req.CertificateChain) == 1 {
+		return nil
+	}
+
+	for _, c := range req.CertificateChain[1:] {
+		if err := validateIntermediateCertificate(c); err != nil {
+			return fmt.Errorf("invalid intermediate certificate provided in chain: %w", err)
+		}
+
+		fp, err := fingerprint.New(c.Raw, crypto.SHA1, fingerprint.HexFingerprint)
+		if err != nil {
+			return fmt.Errorf("failed calculating certificate SHA1 fingerprint: %w", err)
+		}
+
+		if err := k.StoreCertificate(&apiv1.StoreCertificateRequest{
+			Name: uri.New("capi", url.Values{
+				HashArg:                []string{fp},
+				StoreLocationArg:       []string{u.intermediateStoreLocation},
+				StoreNameArg:           []string{u.intermediateStoreName},
+				SkipFindCertificateKey: []string{"true"},
+			}).String(),
+			Certificate: c,
+		}); err != nil {
+			return err
+		}
+	}
+
+	return nil
+}
+
 // DeleteCertificate deletes a certificate from the Windows certificate store. It uses
 // largely the same logic for searching for the certificate as [LoadCertificate], but
 // deletes it as soon as it's found.
@@ -774,37 +908,19 @@ func (k *CAPIKMS) StoreCertificate(req *apiv1.StoreCertificateRequest) error {
 // Notice: This method is EXPERIMENTAL and may be changed or removed in a later
 // release.
 func (k *CAPIKMS) DeleteCertificate(req *apiv1.DeleteCertificateRequest) error {
-	u, err := uri.ParseWithScheme(Scheme, req.Name)
+	u, err := parseURI(req.Name)
 	if err != nil {
-		return fmt.Errorf("failed to parse URI: %w", err)
-	}
-
-	sha1Hash, err := u.GetHexEncoded(HashArg)
-	if err != nil {
-		return fmt.Errorf("failed getting %s from URI %q: %w", HashArg, req.Name, err)
-	}
-	keyID := u.Get(KeyIDArg)
-	issuerName := u.Get(IssuerNameArg)
-	serialNumber := u.Get(SerialNumberArg)
-
-	var storeLocation string
-	if storeLocation = u.Get(StoreLocationArg); storeLocation == "" {
-		storeLocation = UserStore
+		return err
 	}
 
 	var certStoreLocation uint32
-	switch storeLocation {
-	case UserStore:
+	switch u.storeLocation {
+	case UserStoreLocation:
 		certStoreLocation = certStoreCurrentUser
-	case MachineStore:
+	case MachineStoreLocation:
 		certStoreLocation = certStoreLocalMachine
 	default:
-		return fmt.Errorf("invalid cert store location %q", storeLocation)
-	}
-
-	var storeName string
-	if storeName = u.Get(StoreNameArg); storeName == "" {
-		storeName = "My"
+		return fmt.Errorf("invalid cert store location %q", u.storeLocation)
 	}
 
 	st, err := windows.CertOpenStore(
@@ -812,23 +928,23 @@ func (k *CAPIKMS) DeleteCertificate(req *apiv1.DeleteCertificateRequest) error {
 		0,
 		0,
 		certStoreLocation,
-		uintptr(unsafe.Pointer(wide(storeName))))
+		uintptr(unsafe.Pointer(wide(u.storeName))))
 	if err != nil {
-		return fmt.Errorf("CertOpenStore for the %q store %q returned: %w", storeLocation, storeName, err)
+		return fmt.Errorf("CertOpenStore for the %q store %q returned: %w", u.storeLocation, u.storeName, err)
 	}
 
 	var certHandle *windows.CertContext
 
 	switch {
-	case len(sha1Hash) > 0:
-		if len(sha1Hash) != 20 {
-			return fmt.Errorf("decoded %s has length %d; expected 20 bytes for SHA-1", HashArg, len(sha1Hash))
+	case len(u.hash) > 0:
+		if len(u.hash) != 20 {
+			return fmt.Errorf("decoded %s has length %d; expected 20 bytes for SHA-1", HashArg, len(u.hash))
 		}
 		searchData := CERT_ID_KEYIDORHASH{
 			idChoice: CERT_ID_SHA1_HASH,
 			KeyIDOrHash: CRYPTOAPI_BLOB{
-				len:  uint32(len(sha1Hash)),
-				data: uintptr(unsafe.Pointer(&sha1Hash[0])),
+				len:  uint32(len(u.hash)),
+				data: uintptr(unsafe.Pointer(&u.hash[0])),
 			},
 		}
 		certHandle, err = findCertificateInStore(st,
@@ -847,18 +963,12 @@ func (k *CAPIKMS) DeleteCertificate(req *apiv1.DeleteCertificateRequest) error {
 			return fmt.Errorf("failed removing certificate: %w", err)
 		}
 		return nil
-	case keyID != "":
-		keyID = strings.TrimPrefix(keyID, "0x") // Support specifying the hash as 0x like with serial
-
-		keyIDBytes, err := hex.DecodeString(keyID)
-		if err != nil {
-			return fmt.Errorf("%s must be in hex format: %w", KeyIDArg, err)
-		}
+	case len(u.keyID) > 0:
 		searchData := CERT_ID_KEYIDORHASH{
 			idChoice: CERT_ID_KEY_IDENTIFIER,
 			KeyIDOrHash: CRYPTOAPI_BLOB{
-				len:  uint32(len(keyIDBytes)),
-				data: uintptr(unsafe.Pointer(&keyIDBytes[0])),
+				len:  uint32(len(u.keyID)),
+				data: uintptr(unsafe.Pointer(&u.keyID[0])),
 			},
 		}
 		certHandle, err = findCertificateInStore(st,
@@ -877,21 +987,21 @@ func (k *CAPIKMS) DeleteCertificate(req *apiv1.DeleteCertificateRequest) error {
 			return fmt.Errorf("failed removing certificate: %w", err)
 		}
 		return nil
-	case issuerName != "" && serialNumber != "":
+	case u.issuerName != "" && u.serialNumber != "":
 		// TODO: Replace this search with a CERT_ID + CERT_ISSUER_SERIAL_NUMBER search instead
 		// https://learn.microsoft.com/en-us/windows/win32/api/wincrypt/ns-wincrypt-cert_id
 		// https://learn.microsoft.com/en-us/windows/win32/api/wincrypt/ns-wincrypt-cert_issuer_serial_number
 		var serialBytes []byte
-		if strings.HasPrefix(serialNumber, "0x") {
-			serialNumber = strings.TrimPrefix(serialNumber, "0x")
-			serialNumber = strings.TrimPrefix(serialNumber, "00") // Comparison fails if leading 00 is not removed
-			serialBytes, err = hex.DecodeString(serialNumber)
+		if strings.HasPrefix(u.serialNumber, "0x") {
+			u.serialNumber = strings.TrimPrefix(u.serialNumber, "0x")
+			u.serialNumber = strings.TrimPrefix(u.serialNumber, "00") // Comparison fails if leading 00 is not removed
+			serialBytes, err = hex.DecodeString(u.serialNumber)
 			if err != nil {
 				return fmt.Errorf("invalid hex format for %s: %w", SerialNumberArg, err)
 			}
 		} else {
 			bi := new(big.Int)
-			bi, ok := bi.SetString(serialNumber, 10)
+			bi, ok := bi.SetString(u.serialNumber, 10)
 			if !ok {
 				return fmt.Errorf("invalid %s - must be in hex or integer format", SerialNumberArg)
 			}
@@ -903,7 +1013,7 @@ func (k *CAPIKMS) DeleteCertificate(req *apiv1.DeleteCertificateRequest) error {
 				encodingX509ASN|encodingPKCS7,
 				0,
 				findIssuerStr,
-				uintptr(unsafe.Pointer(wide(issuerName))), prevCert)
+				uintptr(unsafe.Pointer(wide(u.issuerName))), prevCert)
 			if err != nil {
 				return fmt.Errorf("findCertificateInStore failed: %w", err)
 			}
@@ -931,18 +1041,18 @@ func (k *CAPIKMS) DeleteCertificate(req *apiv1.DeleteCertificateRequest) error {
 	}
 }
 
-func (k *CAPIKMS) getKeyFlags(u *uri.URI) (uint32, error) {
+func (k *CAPIKMS) getKeyFlags(u *uriAttributes) (uint32, error) {
 	keyFlags := uint32(0)
 
-	switch u.Get(StoreLocationArg) {
-	case MachineStore:
+	switch u.storeLocation {
+	case MachineStoreLocation:
 		if k.providerName == ProviderMSSC {
 			return 0, fmt.Errorf("machine store cannot be used with the %s", ProviderMSSC)
 		}
 
 		keyFlags |= NCRYPT_MACHINE_KEY_FLAG
 
-	case UserStore:
+	case UserStoreLocation:
 		if k.providerName == ProviderMSPCP {
 			return 0, fmt.Errorf("user store cannot be used with the %s", ProviderMSPCP)
 		}
@@ -950,7 +1060,7 @@ func (k *CAPIKMS) getKeyFlags(u *uri.URI) (uint32, error) {
 	case "":
 
 	default:
-		return 0, fmt.Errorf("invalid storeLocation %v", u.Get(StoreLocationArg))
+		return 0, fmt.Errorf("invalid storeLocation %v", u.storeLocation)
 	}
 
 	return keyFlags, nil
@@ -1041,4 +1151,19 @@ func (s *CAPISigner) Public() crypto.PublicKey {
 	return s.PublicKey
 }
 
+func validateIntermediateCertificate(c *x509.Certificate) error {
+	switch {
+	case !c.IsCA:
+		return fmt.Errorf("certificate with serial %q is not a CA certificate", c.SerialNumber.String())
+	case !c.BasicConstraintsValid:
+		return fmt.Errorf("certificate with serial %q has invalid basic constraints", c.SerialNumber.String())
+	case bytes.Equal(c.AuthorityKeyId, c.SubjectKeyId):
+		return fmt.Errorf("certificate with serial %q has equal subject and authority key IDs", c.SerialNumber.String())
+	case c.CheckSignatureFrom(c) == nil:
+		return fmt.Errorf("certificate with serial %q is self-signed root CA", c.SerialNumber.String())
+	}
+
+	return nil
+}
+
 var _ apiv1.CertificateManager = (*CAPIKMS)(nil)
