@@ -2,6 +2,8 @@ package x509util
 
 import (
 	"bytes"
+	"crypto/ed25519"
+	"crypto/rand"
 	"crypto/x509"
 	"crypto/x509/pkix"
 	"encoding/asn1"
@@ -241,6 +243,7 @@ func TestSubjectAlternativeName_Set(t *testing.T) {
 		{"AutoIPAdd", fields{"", "::1"}, args{&x509.Certificate{IPAddresses: []net.IP{net.ParseIP("127.0.0.1")}}}, &x509.Certificate{IPAddresses: []net.IP{net.ParseIP("127.0.0.1"), net.ParseIP("::1")}}},
 		{"AutoURI", fields{"Auto", "https://foo.com"}, args{&x509.Certificate{}}, &x509.Certificate{URIs: []*url.URL{{Scheme: "https", Host: "foo.com"}}}},
 		{"AutoURIAdd", fields{"", "uri:foo:bar"}, args{&x509.Certificate{URIs: []*url.URL{{Scheme: "https", Host: "foo.com"}}}}, &x509.Certificate{URIs: []*url.URL{{Scheme: "https", Host: "foo.com"}, {Scheme: "uri", Opaque: "foo:bar"}}}},
+		{"skipEmpty", fields{"auto", ""}, args{&x509.Certificate{}}, &x509.Certificate{}},
 		{"panic", fields{"panic", "foo.com"}, args{&x509.Certificate{}}, &x509.Certificate{DNSNames: []string{"foo.com"}}},
 		{"panicAdd", fields{"panic", "bar.com"}, args{&x509.Certificate{DNSNames: []string{"foo.com"}}}, &x509.Certificate{DNSNames: []string{"foo.com"}}},
 	}
@@ -517,6 +520,7 @@ func TestKeyUsage_UnmarshalJSON(t *testing.T) {
 		// Normalized
 		{"DigitalSignature", args{[]byte(`"DigitalSignature"`)}, KeyUsage(x509.KeyUsageDigitalSignature), false},
 		{"ContentCommitment", args{[]byte(`"ContentCommitment"`)}, KeyUsage(x509.KeyUsageContentCommitment), false},
+		{"NonRepudiation", args{[]byte(`"NonRepudiation"`)}, KeyUsage(x509.KeyUsageContentCommitment), false},
 		{"KeyEncipherment", args{[]byte(`"KeyEncipherment"`)}, KeyUsage(x509.KeyUsageKeyEncipherment), false},
 		{"DataEncipherment", args{[]byte(`"DataEncipherment"`)}, KeyUsage(x509.KeyUsageDataEncipherment), false},
 		{"KeyAgreement", args{[]byte(`"KeyAgreement"`)}, KeyUsage(x509.KeyUsageKeyAgreement), false},
@@ -527,6 +531,7 @@ func TestKeyUsage_UnmarshalJSON(t *testing.T) {
 		// Snake case
 		{"digital_signature", args{[]byte(`"digital_signature"`)}, KeyUsage(x509.KeyUsageDigitalSignature), false},
 		{"content_commitment", args{[]byte(`"content_commitment"`)}, KeyUsage(x509.KeyUsageContentCommitment), false},
+		{"non_repudiation", args{[]byte(`"non_repudiation"`)}, KeyUsage(x509.KeyUsageContentCommitment), false},
 		{"key_encipherment", args{[]byte(`"key_encipherment"`)}, KeyUsage(x509.KeyUsageKeyEncipherment), false},
 		{"data_encipherment", args{[]byte(`"data_encipherment"`)}, KeyUsage(x509.KeyUsageDataEncipherment), false},
 		{"key_agreement", args{[]byte(`"key_agreement"`)}, KeyUsage(x509.KeyUsageKeyAgreement), false},
@@ -1245,7 +1250,7 @@ func Test_createSubjectAltNameExtension(t *testing.T) {
 		wantErr bool
 	}{
 		{"ok dns", args{Certificate{
-			DNSNames: []string{"foo.com"},
+			DNSNames: []string{"foo.com", ""},
 		}, false}, Extension{
 			ID:       oidExtensionSubjectAltName,
 			Critical: false,
@@ -1259,21 +1264,21 @@ func Test_createSubjectAltNameExtension(t *testing.T) {
 			Value:    append([]byte{0x30, 9, 0x80 | nameTypeDNS, 7}, []byte("foo.com")...),
 		}, false},
 		{"ok email", args{Certificate{
-			EmailAddresses: []string{"bar@foo.com"},
+			EmailAddresses: []string{"bar@foo.com", ""},
 		}, false}, Extension{
 			ID:       oidExtensionSubjectAltName,
 			Critical: false,
 			Value:    append([]byte{0x30, 13, 0x80 | nameTypeEmail, 11}, []byte("bar@foo.com")...),
 		}, false},
 		{"ok uri", args{Certificate{
-			URIs: []*url.URL{{Scheme: "urn", Opaque: "foo:bar"}},
+			URIs: []*url.URL{{Scheme: "urn", Opaque: "foo:bar"}, {}},
 		}, false}, Extension{
 			ID:       oidExtensionSubjectAltName,
 			Critical: false,
 			Value:    append([]byte{0x30, 13, 0x80 | nameTypeURI, 11}, []byte("urn:foo:bar")...),
 		}, false},
 		{"ok ip", args{Certificate{
-			IPAddresses: []net.IP{net.ParseIP("1.2.3.4")},
+			IPAddresses: []net.IP{net.ParseIP("1.2.3.4"), {}},
 		}, false}, Extension{
 			ID:       oidExtensionSubjectAltName,
 			Critical: false,
@@ -1285,6 +1290,7 @@ func Test_createSubjectAltNameExtension(t *testing.T) {
 				{Type: "email", Value: "bar@foo.com"},
 				{Type: "uri", Value: "urn:foo:bar"},
 				{Type: "ip", Value: "1.2.3.4"},
+				{Type: "auto", Value: ""},
 			},
 		}, false}, Extension{
 			ID:       oidExtensionSubjectAltName,
@@ -1318,7 +1324,7 @@ func Test_createSubjectAltNameExtension(t *testing.T) {
 			}, nil),
 		}, false},
 		{"fail dns", args{Certificate{
-			DNSNames: []string{""},
+			DNSNames: []string{"xn--bücher.example.com"},
 		}, false}, Extension{}, true},
 		{"fail email", args{Certificate{
 			EmailAddresses: []string{"nöt@ia5.com"},
@@ -1665,3 +1671,162 @@ func TestPolicyIdentifiers(t *testing.T) {
 		})
 	}
 }
+
+func TestKeyUsage_Extension(t *testing.T) {
+	pub, _, err := ed25519.GenerateKey(rand.Reader)
+	require.NoError(t, err)
+
+	iss, issPriv := createIssuerCertificate(t, "issuer")
+	mustExtension := func(ku x509.KeyUsage) Extension {
+		crt, err := CreateCertificate(&x509.Certificate{
+			Subject:   pkix.Name{CommonName: "leaf"},
+			PublicKey: pub,
+			KeyUsage:  ku,
+		}, iss, pub, issPriv)
+		require.NoError(t, err)
+
+		for _, ext := range crt.Extensions {
+			if ext.Id.Equal(oidExtensionKeyUsage) {
+				return newExtension(ext)
+			}
+		}
+		return Extension{}
+	}
+
+	all := x509.KeyUsageDigitalSignature |
+		x509.KeyUsageContentCommitment |
+		x509.KeyUsageKeyEncipherment |
+		x509.KeyUsageDataEncipherment |
+		x509.KeyUsageKeyAgreement |
+		x509.KeyUsageCertSign |
+		x509.KeyUsageCRLSign |
+		x509.KeyUsageEncipherOnly |
+		x509.KeyUsageDecipherOnly
+
+	tests := []struct {
+		name      string
+		k         KeyUsage
+		want      Extension
+		assertion assert.ErrorAssertionFunc
+	}{
+		{"ca", KeyUsage(x509.KeyUsageCertSign | x509.KeyUsageCRLSign), mustExtension(x509.KeyUsageCertSign | x509.KeyUsageCRLSign), assert.NoError},
+		{"leaf", KeyUsage(x509.KeyUsageDigitalSignature), mustExtension(x509.KeyUsageDigitalSignature), assert.NoError},
+		{"all", KeyUsage(all), mustExtension(all), assert.NoError},
+		{"empty", KeyUsage(0), Extension{}, assert.NoError},
+	}
+	for _, tt := range tests {
+		t.Run(tt.name, func(t *testing.T) {
+			got, err := tt.k.Extension()
+			tt.assertion(t, err)
+			assert.Equal(t, tt.want, got)
+		})
+	}
+}
+
+func TestExtKeyUsage_Extension(t *testing.T) {
+	pub, _, err := ed25519.GenerateKey(rand.Reader)
+	require.NoError(t, err)
+
+	iss, issPriv := createIssuerCertificate(t, "issuer")
+	mustExtension := func(eku []x509.ExtKeyUsage, unhandledEKUs ...asn1.ObjectIdentifier) Extension {
+		crt, err := CreateCertificate(&x509.Certificate{
+			Subject:            pkix.Name{CommonName: "leaf"},
+			PublicKey:          pub,
+			KeyUsage:           x509.KeyUsageDigitalSignature,
+			ExtKeyUsage:        eku,
+			UnknownExtKeyUsage: unhandledEKUs,
+		}, iss, pub, issPriv)
+		require.NoError(t, err)
+
+		for _, ext := range crt.Extensions {
+			if ext.Id.Equal(oidExtensionExtendedKeyUsage) {
+				return newExtension(ext)
+			}
+		}
+		return Extension{}
+	}
+
+	type args struct {
+		unknownUsages UnknownExtKeyUsage
+	}
+	tests := []struct {
+		name      string
+		k         ExtKeyUsage
+		args      args
+		want      Extension
+		assertion assert.ErrorAssertionFunc
+	}{
+		{"ok", ExtKeyUsage{x509.ExtKeyUsageClientAuth, x509.ExtKeyUsageServerAuth}, args{nil}, mustExtension([]x509.ExtKeyUsage{x509.ExtKeyUsageClientAuth, x509.ExtKeyUsageServerAuth}), assert.NoError},
+		{"ok unhandled", ExtKeyUsage{x509.ExtKeyUsageClientAuth}, args{[]asn1.ObjectIdentifier{
+			{1, 2, 3, 4}, {1, 5, 6, 7},
+		}}, mustExtension([]x509.ExtKeyUsage{x509.ExtKeyUsageClientAuth}, asn1.ObjectIdentifier{1, 2, 3, 4}, asn1.ObjectIdentifier{1, 5, 6, 7}), assert.NoError},
+		{"ok unhandled only", ExtKeyUsage{}, args{[]asn1.ObjectIdentifier{
+			{1, 2, 3, 4}, {1, 5, 6, 7},
+		}}, mustExtension([]x509.ExtKeyUsage{}, asn1.ObjectIdentifier{1, 2, 3, 4}, asn1.ObjectIdentifier{1, 5, 6, 7}), assert.NoError},
+		{"empty", ExtKeyUsage{}, args{UnknownExtKeyUsage{}}, Extension{}, assert.NoError},
+		{"empty nil", ExtKeyUsage{}, args{nil}, Extension{}, assert.NoError},
+		{"fail extKeyUsage", ExtKeyUsage{x509.ExtKeyUsage(100)}, args{nil}, Extension{}, assert.Error},
+		{"fail unhandled", ExtKeyUsage{x509.ExtKeyUsageCodeSigning}, args{[]asn1.ObjectIdentifier{{1, 2, 3, 4}, {5, 6, 7, 8}}}, Extension{}, assert.Error},
+	}
+	for _, tt := range tests {
+		t.Run(tt.name, func(t *testing.T) {
+			got, err := tt.k.Extension(tt.args.unknownUsages)
+			tt.assertion(t, err)
+			assert.Equal(t, tt.want, got)
+		})
+	}
+}
+
+func TestBasicConstraints_Extension(t *testing.T) {
+	pub, _, err := ed25519.GenerateKey(rand.Reader)
+	require.NoError(t, err)
+
+	iss, issPriv := createIssuerCertificate(t, "issuer")
+	mustExtension := func(isCA bool, maxPathLen int, maxPathLenZero bool) Extension {
+		crt, err := CreateCertificate(&x509.Certificate{
+			Subject:               pkix.Name{CommonName: "ca"},
+			PublicKey:             pub,
+			KeyUsage:              x509.KeyUsageCertSign | x509.KeyUsageCRLSign,
+			IsCA:                  isCA,
+			MaxPathLen:            maxPathLen,
+			MaxPathLenZero:        maxPathLenZero,
+			BasicConstraintsValid: true,
+		}, iss, pub, issPriv)
+		require.NoError(t, err)
+
+		for _, ext := range crt.Extensions {
+			if ext.Id.Equal(oidExtensionBasicConstraints) {
+				return newExtension(ext)
+			}
+		}
+		return Extension{}
+	}
+
+	type fields struct {
+		IsCA       bool
+		MaxPathLen int
+	}
+	tests := []struct {
+		name      string
+		fields    fields
+		want      Extension
+		assertion assert.ErrorAssertionFunc
+	}{
+		{"ca 1", fields{true, 1}, mustExtension(true, 1, false), assert.NoError},
+		{"ca 0", fields{true, 0}, mustExtension(true, 0, true), assert.NoError},
+		{"ca -1", fields{true, -1}, mustExtension(true, -1, false), assert.NoError},
+		{"ca -2", fields{true, -2}, mustExtension(true, -1, false), assert.NoError},
+		{"no ca", fields{false, 0}, mustExtension(false, 0, false), assert.NoError},
+	}
+	for _, tt := range tests {
+		t.Run(tt.name, func(t *testing.T) {
+			b := BasicConstraints{
+				IsCA:       tt.fields.IsCA,
+				MaxPathLen: tt.fields.MaxPathLen,
+			}
+			got, err := b.Extension()
+			tt.assertion(t, err)
+			assert.Equal(t, tt.want, got)
+		})
+	}
+}
